"""Btw codes used by E-boekhouden.nl."""
from .string_enum import StringEnum


class BtwCode(StringEnum):
    """Btw codes used by E-boekhouden.nl."""

    hoog_verkoop_19 = "HOOG_VERK"  # BTW hoog, verkopen 19%
    hoog_verkoop_21 = "HOOG_VERK_21"  # BTW hoog, verkopen 21%
    laag_verkoop = "LAAG_VERK"  # BTW laag, verkopen voor 2019 6% daarna 9%
    laag_verkoop_9 = "LAAG_VERK_9"  # BTW laag, verkopen 9%
    verlegd_verkoop_9 = "VERL_VERK_L9"  # BTW Verlegd 9% (1e op de btw-aangifte)
    verlegd_verkoop_21 = "VERL_VERK"  # BTW Verlegd 21% (1e op de btw-aangifte)
    afwijkend = "AFW"  # Afwijkend btw-tarief
    buiten_eu_verkoop_0 = "BU_EU_VERK"  # Leveringen naar buiten de EU 0%
    binnen_eu_verkoop_0 = "BI_EU_VERK"  # Goederen naar binnen de EU 0%
    binnen_eu_diensten_0 = "BI_EU_VERK_D"  # Diensten naar binnen de EU 0%
    afstand_verkopen_0 = "AFST_VERK"  # Afstandsverkopen naar binnen de EU 0%
    laag_inkopen = "LAAG_INK"  # BTW laag, inkopen voor 2019 6% daarna 9%
    laag_inkopen_9 = "LAAG_INK_9"  # BTW laag, inkopen 9%
    verlegd_inkopen_9 = "VERL_INK_L9"  # BTW verlegd, laag, inkopen
    hoog_inkopen = "HOOG_INK"  # BTW hoog, inkopen
    hoog_inkopen_21 = "HOOG_INK_21"  # BTW hoog, inkopen 21%
    verlegd_inkopen_21 = "VERL_INK"  # BTW verlegd, hoog, inkopen
    afwijkend_verkoop = "AFW_VERK"  # Afwijkend btw-tarief verkoop
    buiten_eu_inkopen_0 = "BU_EU_INK"  # Leveringen/diensten van buiten de EU 0%
    binnen_eu_inkopen_0 = "BI_EU_INK"  # Leveringen/diensten van binnen de EU 0%
    geen = "GEEN"  # Geen BTW


btw_codes_hoog = [
    BtwCode.hoog_verkoop_19,
    BtwCode.hoog_verkoop_21,
    BtwCode.hoog_inkopen,
    BtwCode.hoog_inkopen_21,
    BtwCode.verlegd_inkopen_21,
]
btw_codes_laag = [
    BtwCode.laag_verkoop,
    BtwCode.laag_verkoop_9,
    BtwCode.verlegd_verkoop_9,
    BtwCode.laag_inkopen,
    BtwCode.laag_inkopen_9,
]
btw_codes_geen = [x for x in BtwCode if x not in btw_codes_hoog + btw_codes_laag]



"""Row in a Mutatie of E-boekhouden.nl."""
from dataclasses import dataclass
from typing import Optional

from zeep.xsd import SkipValue as ZeepXsdSkipValue

from ..constants import BtwCode, btw_codes_hoog, btw_codes_laag, btw_codes_geen


@dataclass
class MutatieRegel:
    """Row in a Mutatie of E-boekhouden.nl."""

    bedrag_invoer: str  # Decimal
    bedrag_exclusief_btw: str  # Decimal
    bedrag_btw: str  # Decimal
    bedrag_inclusief_btw: str  # Decimal
    btw_code: str  # String
    btw_percentage: str  # Decimal
    tegenrekening_code: str  # String
    kostenplaats_id: Optional[str] = None  # Int

    @classmethod
    def from_bedrag(
        cls,
        bedrag_invoer: float,
        btw_code: BtwCode,
        tegenrekening_code: str,
    ) -> "MutatieRegel":
        """Create a MutatieRegel from amount including vat using BtwCode."""
        if btw_code in btw_codes_hoog:
            btw_percentage = 21
        elif btw_code in btw_codes_laag:
            btw_percentage = 9
        elif btw_code in btw_codes_geen:
            btw_percentage = 0
        else:
            raise ValueError(f"Invalid BTW code: {btw_code}")

        bedrag_exclusief_btw = bedrag_invoer / (1 + btw_percentage / 100)
        bedrag_btw = bedrag_invoer - bedrag_exclusief_btw

        return cls(
            bedrag_invoer=f"{bedrag_invoer:.2f}",
            bedrag_exclusief_btw=f"{bedrag_exclusief_btw:.2f}",
            bedrag_btw=f"{bedrag_btw:.2f}",
            bedrag_inclusief_btw=f"{bedrag_invoer:.2f}",
            btw_code=btw_code,
            btw_percentage=str(btw_percentage),
            tegenrekening_code=tegenrekening_code,
        )

    def export(self):
        """Export to structure used in SOAP of E-Boekhouden.nl."""
        return dict(
            BedragInvoer=self.bedrag_invoer,
            BedragExclBTW=self.bedrag_exclusief_btw,
            BedragBTW=self.bedrag_btw,
            BedragInclBTW=self.bedrag_inclusief_btw,
            BTWCode=self.btw_code,
            BTWPercentage=self.btw_percentage,
            TegenrekeningCode=self.tegenrekening_code,
            KostenplaatsID=self.kostenplaats_id or ZeepXsdSkipValue,
        )

    def to_string(self):
        """Return a string representation of this object."""
        return (
            "Mutatieregel:\n"
            "-------------\n"
            f"BedragInvoer         : {self.bedrag_invoer}\n"
            f"Bedrag Exclusief BTW : {self.bedrag_exclusief_btw}\n"
            f"Bedrag BTW           : {self.bedrag_btw}\n"
            f"Bedrag met BTW       : {self.bedrag_inclusief_btw}\n"
            f"Btw code             : {self.btw_code}\n"
            f"Btw percentage       : {self.btw_percentage}\n"
            f"Tegenrekening        : {self.tegenrekening_code}\n"
            f"Kostenplaats ID      : {self.kostenplaats_id}\n"
        )

    def __repr__(self) -> str:  # pragma: no cover
        """Return a string representation of this object."""
        self.to_string()



"""Including or excluding BTW type used by E-boekhouden.nl."""
from .string_enum import StringEnum


class InExBTW(StringEnum):
    """Including or excluding BTW type used by E-boekhouden.nl."""

    inclusief = "IN"  # Inclusief BTW
    exclusief = "EX"  # Exclusief BTW



"""Eboekhouden client."""
from typing import Optional
import os

from zeep import Client as ZeepClient

from .models import MutatieFilter, Mutatie


class EboekhoudenClient:
    """Eboekhouden client."""

    def __init__(
        self,
        username: Optional[str] = None,
        code1: Optional[str] = None,
        code2: Optional[str] = None,
        server_url: Optional[str] = "https://soap.e-boekhouden.nl/soap.asmx?wsdl",
    ):
        """Initialize Eboekhouden client."""
        self._username = username or os.environ.get("EBOEKHOUDEN_USERNAME")
        self._code1 = code1 or os.environ.get("EBOEKHOUDEN_CODE1")
        self._code2 = code2 or os.environ.get("EBOEKHOUDEN_CODE2")
        self._server_url = server_url or os.environ.get("EBOEKHOUDEN_SERVER_URL")

        if self._username is None or self._code1 is None or self._code2 is None:
            raise ValueError("Incomplete Eboekhouden credentials")

        if self._server_url is None or self._server_url == "":
            raise ValueError("Server URL Eboekhouden missing")

        self._client = ZeepClient(self._server_url)

        self.get_session_id()

    def _check_response(self, response) -> None:
        """Check for general response for errors."""
        if "ErrorMsg" not in response:
            raise ValueError("Unexpected response, do you have the correct server URL?")

        if response["ErrorMsg"] is not None:
            if "LastErrorCode" in response["ErrorMsg"]:
                if response["ErrorMsg"]["LastErrorCode"] is not None:
                    print(response["ErrorMsg"])
                    raise ValueError(
                        "Error interfacing with Eboekhouden, correct credentials?",
                    )
            else:
                if response["ErrorMsg"] is not None:
                    print(response["ErrorMsg"])
                    raise ValueError(
                        "Error interfacing with Eboekhouden, correct credentials?",
                    )

    def get_session_id(self):
        """Get session ID."""
        response = self._client.service.OpenSession(
            Username=self._username,
            SecurityCode1=self._code1,
            SecurityCode2=self._code2,
        )
        self._check_response(response)

        self._session_id = response["SessionID"]

    def get_mutaties(
        self,
        mutatie_nummer: Optional[str] = None,
        mutatie_nummer_van: Optional[str] = None,
        mutatie_nummer_totmet: Optional[str] = None,
        factuur_nummer: Optional[str] = None,
        datum_van: Optional[str] = None,
        datum_totmet: Optional[str] = None,
    ) -> list[dict]:
        """Get mutaties."""
        mutatie_filter = MutatieFilter(
            mutatie_nummer=mutatie_nummer,
            mutatie_nummer_van=mutatie_nummer_van,
            mutatie_nummer_totmet=mutatie_nummer_totmet,
            factuur_nummer=factuur_nummer,
            datum_van=datum_van,
            datum_totmet=datum_totmet,
        )

        self.get_session_id()

        mutaties = self._client.service.GetMutaties(
            SessionID=self._session_id,
            SecurityCode2=self._code2,
            cFilter=mutatie_filter.export(),
        )

        self._check_response(mutaties)

        if mutaties["Mutaties"] is None:
            return []

        return mutaties["Mutaties"]["cMutatieList"]

    def add_mutatie(self, mutatie: Mutatie) -> str:
        """Add mutatie."""
        self.get_session_id()

        response = self._client.service.AddMutatie(
            SessionID=self._session_id,
            SecurityCode2=self._code2,
            oMut=mutatie.export(),
        )

        self._check_response(response)

        if response["MutatieNr"] is None:
            raise ValueError("Error adding mutatie")

        return response["MutatieNr"]



"""Mutatie types used by E-boekhouden.nl."""
from .string_enum import StringEnum


class MutatieSoort(StringEnum):
    """Mutatie types used by E-boekhouden.nl."""

    factuur_ontvangen = "FactuurOntvangen"  # Factuur ontvangen
    factuur_verstuurd = "FactuurVerstuurd"  # Factuur verstuurd
    factuurbetaling_ontvangen = "FactuurbetalingOntvangen"  # Factuurbetaling ontvangen
    factuurbetaling_verstuurd = "FactuurbetalingVerstuurd"  # Factuurbetaling verstuurd
    geld_ontvangen = "GeldOntvangen"  # Geld ontvangen
    geld_uitgeven = "GeldUitgegeven"  # Geld uitgegeven
    memoriaal = "Memoriaal"  # Memoriaal



"""Constants for eboekhouden_python."""
from .btw_code import BtwCode, btw_codes_hoog, btw_codes_laag, btw_codes_geen
from .in_ex_btw import InExBTW
from .mutatie_soort import MutatieSoort


__all__ = [
    BtwCode,
    btw_codes_hoog,
    btw_codes_laag,
    btw_codes_geen,
    InExBTW,
    MutatieSoort,
]



"""Simple string enum class."""
from enum import Enum


class StringEnum(str, Enum):
    """String enum class."""

    def __str__(self) -> str:  # pragma: no cover
        """Return the value of the enum."""
        return self.value



"""Model for GetMutaties filter in e-Boekhouden.nl."""
from dataclasses import dataclass
from typing import Optional

from zeep.xsd import SkipValue as ZeepXsdSkipValue


@dataclass
class MutatieFilter:
    """Filter for GetMutaties in e-Boekhouden."""

    mutatie_nummer: Optional[str] = None
    mutatie_nummer_van: Optional[str] = None
    mutatie_nummer_totmet: Optional[str] = None
    factuur_nummer: Optional[str] = None
    datum_van: Optional[str] = None
    datum_totmet: Optional[str] = None

    def export(self):
        """Export to structure used in SOAP of E-Boekhouden.nl."""
        return dict(
            MutatieNr=self.mutatie_nummer or ZeepXsdSkipValue,
            MutatieNrVan=self.mutatie_nummer_van or ZeepXsdSkipValue,
            MutatieNrTm=self.mutatie_nummer_totmet or ZeepXsdSkipValue,
            Factuurnummer=self.factuur_nummer or ZeepXsdSkipValue,
            DatumVan=self.datum_van or ZeepXsdSkipValue,
            DatumTm=self.datum_totmet or ZeepXsdSkipValue,
        )



"""Main init for eboekhouden_python."""
from .eboekhouden_client import EboekhoudenClient
from .constants import (
    BtwCode,
    InExBTW,
    MutatieSoort,
)


__all__ = [
    BtwCode,
    EboekhoudenClient,
    InExBTW,
    MutatieSoort,
]



"""Mutatie model."""
from dataclasses import dataclass
from typing import Optional

from zeep.xsd import SkipValue as ZeepXsdSkipValue

from .mutatie_regel import MutatieRegel


@dataclass
class Mutatie:
    """Mutatie model."""

    soort: str  # String
    datum: str  # String format example: "2023-01-01"
    rekening: str  # String grootboekrekening, e.g. 8000
    relatie_code: str  # String
    factuur_nummer: str  # String
    omschrijving: str  # String
    betalingstermijn: str  # String in days, e.g. 14
    inclusief_exclusief_btw: str  # String
    mutatie_regels: list[MutatieRegel]  # List of MutatieRegel
    mutatie_nummer: Optional[str] = None  # str=99999
    boekstuk: Optional[str] = None  # String
    betalingskenmerk: Optional[str] = None  # String

    def export(self):
        """Export to structure used in SOAP of E-Boekhouden.nl."""
        return dict(
            MutatieNr=self.mutatie_nummer or ZeepXsdSkipValue,
            Soort=self.soort,
            Datum=self.datum,
            Rekening=self.rekening,
            RelatieCode=self.relatie_code,
            Factuurnummer=self.factuur_nummer,
            Boekstuk=self.boekstuk or ZeepXsdSkipValue,
            Omschrijving=self.omschrijving,
            Betalingstermijn=self.betalingstermijn,
            Betalingskenmerk=self.betalingskenmerk or ZeepXsdSkipValue,
            InExBTW=self.inclusief_exclusief_btw,
            MutatieRegels=dict(cMutatieRegel=[x.export() for x in self.mutatie_regels]),
        )

    def to_string(self):
        """Return a string representation of this object."""
        regels = "----------------------\n".join([x.to_string() for x in self.mutatie_regels])
        return (
            "Mutatie:\n"
            "--------\n"
            f"Mutatienummer        : {self.mutatie_nummer}\n"
            f"Soort                : {self.soort}\n"
            f"Rekening             : {self.rekening}\n"
            f"RelatieCode          : {self.relatie_code}\n"
            f"Factuurnummer        : {self.factuur_nummer}\n"
            f"Omschrijving         : {self.omschrijving}\n"
            f"Betalingstermijn     : {self.betalingstermijn}\n"
            f"Inclusief/exclusief  : {self.inclusief_exclusief_btw}\n"
            f"Boekstuk             : {self.boekstuk}\n"
            f"Betalingskenmerk     : {self.betalingskenmerk}\n"
            f"Mutatieregels        : {len(self.mutatie_regels)}\n"
            "----------------------\n"
            f"{regels}"
        )

    def __repr__(self) -> str:  # pragma: no cover
        """Return a string representation of this object."""
        self.to_string()
